#ifndef PID_v1_h
#define PID_v1_h
#define LIBRARY_VERSION 1.2.1

enum modes : uint8_t {OFF, AUTOMATIC, MANUAL, OVERFLOW};

class PID
{

  public:
  //Constants used in some of the functions below
  #define DIRECT  0
  #define REVERSE  1
  #define P_ON_M 0
  #define P_ON_E 1

  //commonly used functions **************************************************************************
    PID(double*, double*, double*,        // * constructor.  links the PID to the Input, Output, and 
        double, double, double, int, int);//   Setpoint.  Initial tuning parameters are also set here.
                                          //   (overload for specifying proportional mode)

    PID(double*, double*, double*,        // * constructor.  links the PID to the Input, Output, and 
        double, double, double, int);     //   Setpoint.  Initial tuning parameters are also set here
    
    void Mode(int Mode);                  // * sets PID to either Manual or Auto
    int CycleMode();
    bool Compute();                       // * performs the PID calculation.  it should be
                                          //   called every time loop() cycles. ON/OFF and
                                          //   calculation frequency can be set using SetMode
                                          //   SetSampleTime respectively
    int ComputeTune();
    void OutputLimits(double, double);    // * clamps the output to a specific range. 0-255 by default, but
                                          //   it's likely the user will want to change this depending on
                                          //   the application
    


  //available but not commonly used functions ********************************************************
    void SetTunings(double, double,       // * While most users will set the tunings once in the 
                    double);              //   constructor, this function gives the user the option
                                          //   of changing tunings during runtime for Adaptive control
    void SetTunings(double, double,       // * overload for specifying proportional mode
                    double, int);             

    void Direction(int);                  // * Sets the Direction, or "Action" of the controller. DIRECT
                                          //   means the output will increase when error is positive. REVERSE
                                          //   means the opposite.  it's very unlikely that this will be needed
                                          //   once it is set in the constructor.
    void SetSampleTime(int);              // * sets the frequency, in Milliseconds, with which 
                                          //   the PID calculation is performed.  default is 100
                                          
  //Display functions ****************************************************************
    double Kp();                          // These functions query the pid for interal values.
    double Ki();                          //  they were created mainly for the pid front-end,
    double Kd();                          // where it's important to know what is actually 
    modes Mode();                         //  inside the PID.
    int Direction();                      //

  // Auto Tune Public
    void Cancel();                        // * Stops the AutoTune 
    
    void OutputStep(double);           // * how far above and below the starting value will the output step?   
    double OutputStep();               // 

    void LookbackSec(int);                // * how far back are we looking to identify peaks
    int LookbackSec();                    //
    
    void NoiseBand(double);            // * the autotune will ignore signal chatter smaller than this value
    double NoiseBand();                //   this should be acurately set
    
    double TunedKp();                     // * once autotune is complete, these functions contain the
    double TunedKi();                     //   computed tuning parameters.  
    double TunedKd();

  private:
    void Initialize();
    
    double kp;                            // * (P)roportional Tuning Parameter
    double ki;                            // * (I)ntegral Tuning Parameter
    double kd;                            // * (D)erivative Tuning Parameter

    int controllerDirection;
    int pOn;

    double *myInput;                      // * Pointers to the Input, Output, and Setpoint variables
    double *myOutput;                     //   This creates a hard link between the variables and the 
    double *mySetpoint;                   //   PID, freeing the user from having to constantly tell us
                                          //   what these values are.  with pointers we'll just know.
              
    unsigned long lastTime;
    double outputSum, lastInput;

    double outMin, outMax;
    bool inAuto, pOnE;
    modes OpMode;
    
    // Autotune stuff
    void FinishUp();
    bool isMax, isMin;
    double noise_band;
    bool autotune_running;
    unsigned long peak1, peak2;
    int SampleTime;
    int nLookBack;
    int peakType;
    double lastInputs[101];
    double peaks[10];
    int peakCount;
    bool atune_peak_change;
    double absMax, absMin;
    double oStep;
    double outputStart;
    double Ku, Pu;
};
#endif

